import 'package:flutter/material.dart';
import 'package:flutter_riverpod/flutter_riverpod.dart';

class TodoAsyncValueExample extends StatelessWidget {
  const TodoAsyncValueExample({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return const ProviderScope(
      child: TodoList(),
    );
  }
}

class TodoList extends ConsumerStatefulWidget {
  const TodoList({Key? key}) : super(key: key);

  @override
  ConsumerState<TodoList> createState() => _TodoListState();
}

class _TodoListState extends ConsumerState<TodoList> {
  bool _isUncheckedFilter = false;

  @override
  Widget build(BuildContext context) {
    final asyncTodos = ref.watch(todosProvider.selectTodos(_isUncheckedFilter));

    return Column(
      children: [
        TodoListFilter(
          filter: _isUncheckedFilter,
          onFilterChanged: _onFilterChanged,
        ),
        const Divider(height: 1),
        const Padding(
          padding: EdgeInsets.only(top: 16, bottom: 8),
          child: TodoListHeader(),
        ),
        Expanded(
          child: asyncTodos.when(
            data: (todos) => ListView.builder(
              itemBuilder: (_, index) => TodoListItem(
                todos[index],
                onCheckedChanged: (value) => _onCheckedChanged(todos[index], value),
              ),
              itemCount: todos.length,
            ),
            error: (error, _) => Center(
              child: Text(error.toString()),
            ),
            loading: () => const Center(
              child: CircularProgressIndicator(),
            ),
          ),
        ),
      ],
    );
  }

  void _onFilterChanged(bool value) {
    setState(() => _isUncheckedFilter = value);
  }

  void _onCheckedChanged(Todo todo, bool? value) {
    final updatedTodo = todo.copyWith(checked: value != null && value);
    ref.read(todosProvider.notifier).updateTodo(updatedTodo);
  }
}

class TodoListFilter extends StatelessWidget {
  const TodoListFilter({
    Key? key,
    required this.filter,
    required this.onFilterChanged,
  }) : super(key: key);

  final bool filter;
  final ValueChanged<bool> onFilterChanged;

  @override
  Widget build(BuildContext context) {
    return Padding(
      padding: const EdgeInsets.symmetric(horizontal: 16),
      child: Row(
        mainAxisAlignment: MainAxisAlignment.spaceBetween,
        children: [
          const Text("Unchecked"),
          Switch(
            value: filter,
            onChanged: onFilterChanged,
          ),
        ],
      ),
    );
  }
}

class TodoListHeader extends ConsumerWidget {
  const TodoListHeader({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context, WidgetRef ref) {
    final asyncTodosCount = ref.watch(todosCountProvider);

    return Container(
      height: 16,
      alignment: Alignment.center,
      child: asyncTodosCount.when(
        data: (count) => Text(
          "$count todos",
          style: Theme.of(context).textTheme.subtitle2,
        ),
        error: (error, _) => Text(error.toString()),
        loading: () => const AspectRatio(
          aspectRatio: 1,
          child: CircularProgressIndicator(),
        ),
      ),
    );
  }
}

class TodoListItem extends StatelessWidget {
  const TodoListItem(
    this.todo, {
    Key? key,
    this.onCheckedChanged,
  }) : super(key: key);

  final Todo todo;
  final ValueChanged<bool?>? onCheckedChanged;

  @override
  Widget build(BuildContext context) {
    return CheckboxListTile(
      value: todo.checked,
      onChanged: onCheckedChanged,
      title: Text(
        todo.title,
        overflow: TextOverflow.ellipsis,
      ),
      controlAffinity: ListTileControlAffinity.leading,
    );
  }
}

final todosProvider = StateNotifierProvider<TodosStateNotifier, AsyncValue<List<Todo>>>(//
    (ref) => TodosStateNotifier(ref.watch(todosRepositoryProvider)));

class TodosStateNotifier extends StateNotifier<AsyncValue<List<Todo>>> {
  TodosStateNotifier(this.todosRepository) : super(const AsyncValue.loading()) {
    load();
  }

  final TodosRepository todosRepository;

  void load() async {
    final todos = await AsyncValue.guard(() => todosRepository.getTodos());
    if (mounted) {
      state = todos;
    }
  }

  void updateTodo(Todo updatedTodo) async {
    state = const AsyncValue.loading();

    final todos = await AsyncValue.guard(() => todosRepository.updateTodos(updatedTodo));
    if (mounted) {
      state = todos;
    }
  }
}

final todosRepositoryProvider = Provider<TodosRepository>((ref) => FakeTodosRepository(todos));

final todosCountProvider = FutureProvider<int>((ref) => ref.watch(todosRepositoryProvider).getTodosCount());

abstract class TodosRepository {
  Future<List<Todo>> getTodos();

  Future<List<Todo>> updateTodos(Todo updatedTodo);

  Future<int> getTodosCount();
}

class FakeTodosRepository implements TodosRepository {
  FakeTodosRepository(this.todos);

  List<Todo> todos;

  @override
  Future<List<Todo>> getTodos() {
    return Future.delayed(const Duration(seconds: 5), () => todos);
  }

  @override
  Future<List<Todo>> updateTodos(Todo updatedTodo) {
    return Future.delayed(const Duration(seconds: 1), () => todos = todos.copyWithTodo(updatedTodo));
  }

  @override
  Future<int> getTodosCount() {
    return Future.delayed(const Duration(seconds: 3), () => todos.length);
  }
}

const todos = [
  Todo(
    id: 0,
    title: 'Unit test passed',
    checked: true,
  ),
  Todo(
    id: 1,
    title: 'Code reviewed',
    checked: true,
  ),
  Todo(
    id: 2,
    title: 'Acceptance criteria for each issue met',
  ),
  Todo(
    id: 3,
    title: 'Functional tests passed',
  ),
  Todo(
    id: 4,
    title: 'Non-functional requirements met',
  ),
  Todo(
    id: 5,
    title: 'Product owner accepts the User Story',
  ),
];

class Todo {
  const Todo({
    required this.id,
    required this.title,
    this.checked = false,
  });

  final int id;
  final String title;
  final bool checked;

  Todo copyWith({
    int? id,
    String? title,
    bool? checked,
  }) {
    return Todo(
      id: id ?? this.id,
      title: title ?? this.title,
      checked: checked ?? this.checked,
    );
  }
}

extension TodoListExtension on List<Todo> {
  List<Todo> whereTodos(bool isUnchecked) => //
      where((todo) => !isUnchecked || !todo.checked).toList();

  List<Todo> copyWithTodo(Todo updatedTodo) => //
      [for (Todo todo in this) todo.id == updatedTodo.id ? updatedTodo : todo];
}

extension TodoStateNotifierProviderExtension on StateNotifierProvider<TodosStateNotifier, AsyncValue<List<Todo>>> {
  AlwaysAliveProviderListenable<AsyncValue<List<Todo>>> selectTodos(bool isUncheckedFilter) => //
      select((async) => async
          .unwrapPrevious() //
          .whenData((value) => value.whereTodos(isUncheckedFilter)));
}
